# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.8.4.dev0+g920b071.d20230506'
__version_tuple__ = version_tuple = (0, 8, 4, 'dev0', 'g920b071.d20230506')
